from math import sqrt, floor
from pathlib import Path
from typing import List, Union
from uuid import uuid4

import numpy as np
import pandas as pd
from pydantic import BaseModel
from scipy.stats import moment
from skimage.filters import sobel

from skimage.measure import label, regionprops_table, shannon_entropy, find_contours
from sklearn.preprocessing import PolynomialFeatures
from sklearn.linear_model import LinearRegression

from model_server.base.accessors import GenericImageDataAccessor, InMemoryDataAccessor, write_accessor_data_to_file
from model_server.base.models import InstanceSegmentationModel
from model_server.base.process import pad, rescale, resample_to_8bit, make_rgb
from model_server.base.annotators import draw_box_on_patch, draw_contours_on_patch, draw_boxes_on_3d_image
from model_server.base.accessors import PatchStack
from model_server.base.process import mask_largest_object


class PatchParams(BaseModel):
    draw_bounding_box: bool = False
    draw_contour: bool = False
    draw_mask: bool = False
    rescale_clip: float = 0.001
    focus_metric: str = 'max_sobel'
    rgb_overlay_channels: List[Union[int, None]] = [None, None, None]
    rgb_overlay_weights: List[float] = [1.0, 1.0, 1.0]
    pad_to: int = 256


class AnnotatedZStackParams(BaseModel):
    draw_label: bool = False
    channel: int = None


class RoiFilterRange(BaseModel):
    min: float
    max: float


class RoiFilter(BaseModel):
    area: Union[RoiFilterRange, None] = None


class RoiSetMetaParams(BaseModel):
    filters: Union[RoiFilter, None] = None
    expand_box_by: List[int] = [128, 0]


class RoiSetExportParams(BaseModel):
    pixel_probabilities: bool = False
    patches_3d: Union[PatchParams, None] = None
    annotated_patches_2d: Union[PatchParams, None] = None
    patches_2d: Union[PatchParams, None] = None
    annotated_zstacks: Union[AnnotatedZStackParams, None] = None
    object_classes: bool = False



def _get_label_ids(acc_seg_mask: GenericImageDataAccessor, allow_3d=False, connect_3d=True) -> InMemoryDataAccessor:
    """
    Convert binary segmentation mask into either a 2D or 3D object identities map
    :param acc_seg_mask: binary segmentation mask (mono) of either two or three dimensions
    :param allow_3d: return a 3D map if True; return a 2D map of the mask's maximum intensity project if False
    :param connect_3d: objects can span multiple z-positions if True; objects are unique to a single z if False
    :return: object identities map
    """
    if allow_3d and connect_3d:
        nda_la = label(
            acc_seg_mask.data[:, :, 0, :]
        ).astype('uint16')
        return InMemoryDataAccessor(np.expand_dims(nda_la, 2))
    elif allow_3d and not connect_3d:
        nla = 0
        la_3d = np.zeros((*acc_seg_mask.hw, 1, acc_seg_mask.nz), dtype='uint16')
        for zi in range(0, acc_seg_mask.nz):
            la_2d = label(acc_seg_mask.data[:, :, 0, zi]).astype('uint16')
            la_2d[la_2d > 0] = la_2d[la_2d > 0] + nla
            nla = la_2d.max()
            la_3d[:, :, 0, zi] = la_2d
        return InMemoryDataAccessor(la_3d)
    else:
        return InMemoryDataAccessor(
            label(
                acc_seg_mask.data[:, :, 0, :].max(axis=-1)
            ).astype('uint16')
        )


def _focus_metrics():
    return {
        'max_intensity': lambda x: np.max(x),
        'stdev': lambda x: np.std(x),
        'max_sobel': lambda x: np.max(sobel(x)),
        'rms_sobel': lambda x: sqrt(np.mean(sobel(x) ** 2)),
        'entropy': lambda x: shannon_entropy(x),
        'moment': lambda x: moment(x.flatten(), moment=2),
    }


def _safe_add(a, g, b):
    assert a.dtype == b.dtype
    assert a.shape == b.shape
    assert g >= 0.0

    return np.clip(
        a.astype('uint32') + g * b.astype('uint32'),
        0,
        np.iinfo(a.dtype).max
    ).astype(a.dtype)


class RoiSet(object):

    def __init__(
            self,
            acc_raw: GenericImageDataAccessor,
            acc_obj_ids: GenericImageDataAccessor,
            params: RoiSetMetaParams = RoiSetMetaParams(),
    ):
        """
        A set of regions of interest, referenced by their positions and contours in the YXCZ space of stack acc_raw.
        RoiSet contains their internal state, which may be exported as patches, maps, and other products by export methods.
        :param acc_raw: accessor to a generally a multichannel z-stack
        :param acc_obj_ids: accessor to a 2D single-channel object identities map, where each pixel's intensity
            labels its membership in a connected object
        :param params: optional arguments that influence the definition and representation of ROIs
        """
        assert acc_obj_ids.chroma == 1
        self.acc_obj_ids = acc_obj_ids
        self.acc_raw = acc_raw
        self.params = params

        self._df = self.filter_df(
            self.make_df(
                self.acc_raw, self.acc_obj_ids, expand_box_by=params.expand_box_by
            ),
            params.filters,
        )

        self.count = len(self._df)
        self.object_class_maps = {}  # classification results

    def __iter__(self):
        """Expose ROI meta information via the Pandas.DataFrame API"""
        return self._df.itertuples(name='Roi')

    @staticmethod
    def make_df(acc_raw, acc_obj_ids, expand_box_by) -> pd.DataFrame:
        """
        Build dataframe associate object IDs with summary stats
        :param acc_raw: accessor to raw image data
        :param acc_obj_ids: accessor to map of object IDs
        :param expand_box_by: number of pixels to expand bounding box in all directions (without exceeding image boundary)
        # :param deproject: assign object's z-position based on argmax of raw data if True
        :return: pd.DataFrame
        """
        # build dataframe of objects, assign z index to each object

        if acc_obj_ids.nz == 1:  # deproject objects' z-coordinates from argmax of raw image
            df = pd.DataFrame(regionprops_table(
                acc_obj_ids.data[:, :, 0, 0],
                intensity_image=acc_raw.data.argmax(axis=3, keepdims=True)[:, :, 0, 0].astype('uint16'),
                properties=('label', 'area', 'intensity_mean', 'bbox', 'centroid')
            )).rename(columns={'bbox-0': 'y0', 'bbox-1': 'x0', 'bbox-2': 'y1', 'bbox-3': 'x1'})
            df['zi'] = df['intensity_mean'].round().astype('int')

        else:  # objects' z-coordinates come from arg of max count in object identities map
            df = pd.DataFrame(regionprops_table(
                acc_obj_ids.data[:, :, 0, :],
                properties=('label', 'area', 'bbox', 'centroid')
            )).rename(columns={'bbox-0': 'y0', 'bbox-1': 'x0', 'bbox-2': 'z0', 'bbox-3': 'y1', 'bbox-4': 'x1', 'bbox-5': 'z1'})
            df['zi'] = df['label'].apply(lambda x: (acc_obj_ids.data == x).sum(axis=(0, 1, 2)).argmax())

        # compute expanded bounding boxes
        h, w, c, nz = acc_raw.shape
        ebxy, ebz = expand_box_by
        df['ebb_y0'] = (df.y0 - ebxy).apply(lambda x: max(x, 0))
        df['ebb_y1'] = (df.y1 + ebxy).apply(lambda x: min(x, h))
        df['ebb_x0'] = (df.x0 - ebxy).apply(lambda x: max(x, 0))
        df['ebb_x1'] = (df.x1 + ebxy).apply(lambda x: min(x, w))
        df['ebb_z0'] = (df.zi - ebz).apply(lambda x: max(x, 0))
        df['ebb_z1'] = (df.zi + ebz).apply(lambda x: min(x, nz))
        df['ebb_h'] = df['ebb_y1'] - df['ebb_y0']
        df['ebb_w'] = df['ebb_x1'] - df['ebb_x0']
        df['ebb_nz'] = df['ebb_z1'] - df['ebb_z0'] + 1

        # compute relative bounding boxes
        df['rel_y0'] = df.y0 - df.ebb_y0
        df['rel_y1'] = df.y1 - df.ebb_y0
        df['rel_x0'] = df.x0 - df.ebb_x0
        df['rel_x1'] = df.x1 - df.ebb_x0

        assert np.all(df['rel_x1'] <= (df['ebb_x1'] - df['ebb_x0']))
        assert np.all(df['rel_y1'] <= (df['ebb_y1'] - df['ebb_y0']))

        df['slice'] = df.apply(
            lambda r:
            np.s_[int(r.y0): int(r.y1), int(r.x0): int(r.x1), :, int(r.zi): int(r.zi + 1)],
            axis=1
        )
        df['expanded_slice'] = df.apply(
            lambda r:
            np.s_[int(r.ebb_y0): int(r.ebb_y1), int(r.ebb_x0): int(r.ebb_x1), :, int(r.ebb_z0): int(r.ebb_z1) + 1],
            axis=1
        )
        df['relative_slice'] = df.apply(
            lambda r:
            np.s_[int(r.rel_y0): int(r.rel_y1), int(r.rel_x0): int(r.rel_x1), :, :],
            axis=1
        )
        df['binary_mask'] = df.apply(
            lambda r: (acc_obj_ids.data == r.label)[r.y0: r.y1, r.x0: r.x1, 0, 0],
            axis=1
        )
        return df


    @staticmethod
    def from_df_and_patch_masks(
            acc_raw: GenericImageDataAccessor,
            df: pd.DataFrame,
            patch_masks: dict, # dict of ndarray, where key is integer label
            params: RoiSetMetaParams = RoiSetMetaParams(),
    ):
        assert len(df) == len(patch_masks)
        zmask = np.zeros((*acc_raw.hw, 1, acc_raw.nz), dtype=bool)

        def _label_obj(r):
            # sl = np.s_[r.ebb_y0:r.ebb_y1, r.ebb_x0:r.ebb_x1, :, r.zi:r.zi + 1]
            sl = np.s_[r.y0:r.y1, r.x0:r.x1, :, r.zi:r.zi + 1]
            rsl = np.s_[r.rel_y0:r.rel_y1, r.rel_x0:r.rel_x1, :, :]
            try:
                zmask[sl] = patch_masks[r.label][rsl]
                print(f'ebb (h, w) = {r.ebb_h, r.ebb_w}')
            except ValueError as ve:
                print(ve)

        df.apply(lambda x: _label_obj(x), axis=1)
        obj_ids = _get_label_ids(zmask, allow_3d=True, connect_3d=False)
        return RoiSet(acc_raw, obj_ids, params)


    @staticmethod
    def filter_df(df: pd.DataFrame, filters: RoiFilter = None) -> pd.DataFrame:
        query_str = 'label > 0'  # always true
        if filters is not None:  # parse filters
            for k, val in filters.dict(exclude_unset=True).items():
                assert k in ('area')
                vmin = val['min']
                vmax = val['max']
                assert vmin >= 0
                query_str = query_str + f' & {k} > {vmin} & {k} < {vmax}'
        return df.loc[df.query(query_str).index, :]

    def get_df(self) -> pd.DataFrame:
        return self._df

    def get_slices(self) -> pd.Series:
        return self.get_df()['slice']

    def add_df_col(self, name, se: pd.Series) -> None:
        self._df[name] = se

    def get_multichannel_projection(self):
        if self.count:
            projected = project_stack_from_focal_points(
                self._df['centroid-0'].to_numpy(),
                self._df['centroid-1'].to_numpy(),
                self._df['zi'].to_numpy(),
                self.acc_raw,
                degree=4,
            )
        else:  # else just return MIP
            projected = self.acc_raw.data.max(axis=-1)
        return projected

    def get_patches_acc(self, channel=None, **kwargs):  # padded, un-annotated 2d patches
        if channel:
            patches_df = self.get_patches(white_channel=channel, **kwargs)
        else:
            patches_df = self.get_patches(**kwargs)
        return PatchStack(list(patches_df.patch))

    def export_annotated_zstack(self, where, prefix='zstack', **kwargs) -> Path:
        annotated = InMemoryDataAccessor(draw_boxes_on_3d_image(self, **kwargs))
        fp = where / (prefix + '.tif')
        write_accessor_data_to_file(fp, annotated)
        return fp

    def get_zmask(self, mask_type='boxes'):
        """
        Return a mask of same dimensionality as raw data

        :param kwargs: variable-length keyword arguments
            mask_type: if 'boxes', zmask is True in each object's complete bounding box; otherwise 'contours'
        """

        assert mask_type in ('contours', 'boxes')
        zi_st = np.zeros(self.acc_raw.shape, dtype='bool')
        lamap = self.acc_obj_ids.data

        # make an object map where label is replaced by focus position in stack and background is -1
        lut = np.zeros(lamap.max() + 1) - 1
        df = self.get_df()
        lut[df.label] = df.zi

        if mask_type == 'contours':
            zi_map = (lut[lamap] + 1.0).astype('int')
            idxs = np.array(zi_map) - 1
            np.put_along_axis(
                zi_st,
                np.expand_dims(idxs, (2, 3)),
                1,
                axis=3
            )

            # change background level from to 0 in final frame
            zi_st[:, :, :, -1][lamap == 0] = 0

        elif mask_type == 'boxes':
            for roi in self:
                zi_st[roi.relative_slice] = 1

        return zi_st


    def classify_by(self, name: str, channel: int, object_classification_model: InstanceSegmentationModel, ):

        # do this on a patch basis, i.e. only one object per frame
        obmap_patches = object_classification_model.label_patch_stack(
            self.get_patches_acc(channel=channel, expaned=False, pad_to=None),
            self.get_patch_masks_acc(expanded=False, pad_to=None)
        )

        om = np.zeros(self.acc_obj_ids.shape, self.acc_obj_ids.dtype)

        self._df['classify_by_' + name] = pd.Series(dtype='Int64')

        # assign labels to object map:
        for i, roi in enumerate(self):
            oc = np.unique(
                mask_largest_object(
                    obmap_patches.iat(i).data
                )
            )[1]
            self._df.loc[roi.Index, 'classify_by_' + name] = oc
            om[self.acc_obj_ids.data == roi.label] = oc
        self.object_class_maps[name] = InMemoryDataAccessor(om)

    def export_dataframe(self, csv_path: Path):
        csv_path.parent.mkdir(parents=True, exist_ok=True)
        self._df.drop(['expanded_slice', 'slice', 'relative_slice', 'binary_mask'], axis=1).to_csv(csv_path, index=False)
        return csv_path


    def export_patch_masks(self, where: Path, pad_to: int = None, prefix='mask', expanded=False) -> list:
        patches_acc = self.get_patch_masks_acc(pad_to=pad_to, expanded=expanded)

        exported = []
        for i, roi in enumerate(self):  # assumes index of patches_acc is same as dataframe
            patch = patches_acc.iat_yxcz(i)
            ext = 'png'
            fname = f'{prefix}-la{roi.label:04d}-zi{roi.zi:04d}.{ext}'
            write_accessor_data_to_file(where / fname, patch)
            exported.append(fname)
        return exported


    def export_patches(self, where: Path, prefix='patch', **kwargs) -> list:
        make_3d = kwargs.get('make_3d', False)
        patches_df = self.get_patches(**kwargs)

        def _export_patch(roi):
            patch = InMemoryDataAccessor(roi.patch)
            ext = 'tif' if make_3d or patch.chroma > 3 else 'png'
            fname = f'{prefix}-la{roi.label:04d}-zi{roi.zi:04d}.{ext}'

            if patch.dtype is np.dtype('uint16'):
                resampled = InMemoryDataAccessor(resample_to_8bit(patch.data))
                write_accessor_data_to_file(where / fname, resampled)
            else:
                write_accessor_data_to_file(where / fname, patch)

            exported.append(where / fname)

        exported = []
        for roi in patches_df.itertuples():  # just used for label info
            _export_patch(roi)

        return exported

    def get_patch_masks(self, pad_to: int = None, expanded: bool = False) -> pd.DataFrame:
        def _make_patch_mask(roi):
            if expanded:
                patch = np.zeros((roi.ebb_h, roi.ebb_w, 1, 1), dtype='uint8')
                patch[roi.relative_slice][:, :, 0, 0] = roi.binary_mask * 255
            else:
                patch = np.zeros((roi.y1 - roi.y0, roi.x1 - roi.x0, 1, 1), dtype='uint8')
                patch[:, :, 0, 0] = roi.binary_mask * 255

            if pad_to:
                patch = pad(patch, pad_to)
            return patch

        dfe = self._df.copy()
        dfe['patch_mask'] = dfe.apply(lambda r: _make_patch_mask(r), axis=1)
        return dfe

    def get_patch_masks_acc(self, **kwargs) -> PatchStack:
        return PatchStack(list(self.get_patch_masks(**kwargs).patch_mask))

    def get_patches(
            self,
            rescale_clip: float = 0.0,
            pad_to: int = None,
            make_3d: bool = False,
            focus_metric: str = None,
            rgb_overlay_channels: list = None,
            rgb_overlay_weights: list = [1.0, 1.0, 1.0],
            white_channel: int = None,
            expanded=False,
            **kwargs
    ) -> pd.DataFrame:

        # arrange RGB channels if so specified, otherwise copy roiset.raw_acc data
        raw = self.acc_raw
        if isinstance(rgb_overlay_channels, (list, tuple)) and isinstance(rgb_overlay_weights, (list, tuple)):
            assert all([c < raw.chroma for c in rgb_overlay_channels if c is not None])
            assert len(rgb_overlay_channels) == 3
            assert len(rgb_overlay_weights) == 3

            if white_channel:
                assert white_channel < raw.chroma
                stack = raw.data[:, :, [white_channel, white_channel, white_channel], :]
            else:
                stack = np.zeros([*raw.shape[0:2], 3, raw.shape[3]], dtype=raw.dtype)

            for ii, ci in enumerate(rgb_overlay_channels):
                if ci is None:
                    continue
                assert isinstance(ci, int)
                assert ci < raw.chroma
                stack[:, :, ii, :] = _safe_add(
                    stack[:, :, ii, :],  # either black or grayscale channel
                    rgb_overlay_weights[ii],
                    raw.data[:, :, ci, :]
                )
        else:
            if white_channel:  # interpret as just a single channel
                assert white_channel < raw.chroma
                annotate_rgb = False
                for k in ['contour_channel', 'bounding_box_channel', 'mask_channel']:
                    ca = kwargs.get(k)
                    if ca is None:
                        continue
                    assert (ca < raw.chroma)
                    if ca != white_channel:
                        annotate_rgb = True
                        break
                if annotate_rgb:  # make RGB patches anyway to include annotation color
                    stack = raw.data[:, :, [white_channel, white_channel, white_channel], :]
                else:  # make monochrome patches
                    stack = raw.data[:, :, [white_channel], :]
            else:
                stack = raw.data

        def _make_patch(roi):
            if expanded:
                patch3d = stack[roi.expanded_slice]
                subpatch = patch3d[roi.relative_slice]
            else:
                patch3d = stack[roi.slice]
                subpatch = patch3d

            ph, pw, pc, pz = patch3d.shape

            # make a 3d patch
            if make_3d or not expanded:
                patch = patch3d

            # make a 2d patch, find optimal z-position determined by focus_metric function on each channel separately
            elif focus_metric is not None:
                foc = _focus_metrics()[focus_metric]

                patch = np.zeros([ph, pw, pc, 1], dtype=patch3d.dtype)

                for ci in range(0, pc):
                    me = [foc(subpatch[:, :, ci, zi]) for zi in range(0, pz)]
                    zif = np.argmax(me)
                    patch[:, :, ci, 0] = patch3d[:, :, ci, zif]

            # make a 2d patch from middle of z-stack
            else:
                zim = floor(pz / 2)
                patch = patch3d[:, :, :, [zim]]

            assert len(patch.shape) == 4

            mask = np.zeros(patch3d.shape[0:2], dtype=bool)
            if expanded:
                mask[roi.relative_slice[0:2]] = roi.binary_mask
            else:
                mask = roi.binary_mask

            if rescale_clip is not None:
                patch = rescale(patch, rescale_clip)

            if kwargs.get('draw_bounding_box') is True and expanded:
                bci = kwargs.get('bounding_box_channel', 0)
                assert bci < 3
                if bci > 0:
                    patch = make_rgb(patch)
                for zi in range(0, patch.shape[3]):
                    patch[:, :, bci, zi] = draw_box_on_patch(
                        patch[:, :, bci, zi],
                        ((roi.rel_x0, roi.rel_y0), (roi.rel_x1, roi.rel_y1)),
                        linewidth=kwargs.get('bounding_box_linewidth', 1)
                    )

            if kwargs.get('draw_mask'):
                mci = kwargs.get('mask_channel', 0)
                # mask = np.zeros(patch.shape[0:2], dtype=bool)
                # mask[roi.relative_slice[0:2]] = roi.binary_mask
                for zi in range(0, patch.shape[3]):
                    patch[:, :, mci, zi] = np.invert(mask) * patch[:, :, mci, zi]

            if kwargs.get('draw_contour'):
                mci = kwargs.get('contour_channel', 0)
                # mask = np.zeros(patch.shape[0:2], dtype=bool)
                # mask[roi.relative_slice[0:2]] = roi.binary_mask

                for zi in range(0, patch.shape[3]):
                    patch[:, :, mci, zi] = draw_contours_on_patch(
                        patch[:, :, mci, zi],
                        find_contours(mask)
                    )

            if pad_to and expanded:
                patch = pad(patch, pad_to)
            return patch

        dfe = self._df.copy()
        dfe['patch'] = dfe.apply(lambda r: _make_patch(r), axis=1)
        return dfe

    def run_exports(self, where: Path, channel, prefix, params: RoiSetExportParams) -> dict:
        """
        Export various representations of ROIs, e.g. patches, annotated stacks, and object maps.
        :param where: path of directory in which to write all export products
        :param channel: color channel of products to export
        :param prefix: prefix of the name of each product's file or subfolder
        :param params: RoiSetExportParams object describing which products to export and with which parameters
        :return: dict of Path objects describing the location of single-file export products
        """
        record = {}
        if not self.count:
            return

        # export dataframe and patch masks
        record['dataframe'] = self.export_dataframe(where / 'dataframe' / (prefix + '.csv'))
        self.export_patch_masks(where / 'patch_masks', prefix=prefix, pad_to=None)

        for k in params.dict().keys():
            subdir = where / k
            pr = prefix
            kp = params.dict()[k]
            if kp is None:
                continue
            if k == 'patches_3d':
                files = self.export_patches(
                    subdir, white_channel=channel, prefix=pr, make_3d=True, **kp
                )
            if k == 'annotated_patches_2d':
                files = self.export_patches(
                    subdir, prefix=pr, make_3d=False, white_channel=channel,
                    bounding_box_channel=1, bounding_box_linewidth=2, **kp,
                )
            if k == 'patches_2d':
                files = self.export_patches(
                    subdir, white_channel=channel, prefix=pr, make_3d=False, **kp
                )
                df_patches = pd.DataFrame(files)
                self._df = pd.merge(self._df, df_patches, left_index=True, right_on='df_index').drop(columns='df_index')
                self._df['patch_id'] = self._df.apply(lambda _: uuid4(), axis=1)
            if k == 'annotated_zstacks':
                self.export_annotated_zstack(subdir, prefix=pr, **kp)
            if k == 'object_classes':
                for kc, acc in self.object_class_maps.items():
                    fp = subdir / kc / (pr + '.tif')
                    write_accessor_data_to_file(fp, acc)
                    record[f'{k}_{kc}'] = fp

        return record


def project_stack_from_focal_points(
        xx: np.ndarray,
        yy: np.ndarray,
        zz: np.ndarray,
        stack: GenericImageDataAccessor,
        degree: int = 2,
) -> np.ndarray:
    """
    Given a set of 3D points, project a multichannel z-stack based on a surface fit of the provided points
    :param xx: vector of point x-coordinates
    :param yy: vector of point y-coordinates
    :param zz: vector of point z-coordinates
    :param stack: z-stack to project
    :param degree: order of polynomial to fit
    :return: multichannel 2d projected image array
    """
    assert xx.shape == yy.shape
    assert xx.shape == zz.shape

    poly = PolynomialFeatures(degree=degree)
    X = np.stack([xx, yy]).T
    features = poly.fit_transform(X, zz)
    model = LinearRegression(fit_intercept=False)
    model.fit(features, zz)

    xy_indices = np.indices(stack.hw).reshape(2, -1).T
    xy_features = np.dot(
        poly.fit_transform(xy_indices, zz),
        model.coef_
    )
    zi_image = xy_features.reshape(
        stack.hw
    ).round().clip(
        0, (stack.nz - 1)
    ).astype('uint16')

    return np.take_along_axis(
        stack.data,
        np.repeat(
            np.expand_dims(zi_image, (2, 3)),
            stack.chroma,
            axis=2
        ),
        axis=3
    )


